""" 
Script to import tuning and readout id channel mapping, and detector calibration
information into manifest dbs for book loading.  At present this just works in
the configuration where it has access to both level 2 and level 3 indexing. This
is technically possible with just level 3 data / indexing but requires some
still non-existant tools. 

Configuration file required::

    config = {
        'archive': {
            'detset': {
                'root_dir': /path/to/detset/root,
                'index': 'detset.sqlite',
                'h5file': 'detset.h5',
                'context': 'context.yaml',
                'write_relpath': True
            },
            'det_cal': {
                'root_dir': /path/to/det_cal/root,
                'index': 'det_cal.sqlite',
                'h5file': 'det_cal.h5,
                'context': 'context.yaml',
                'failed_obsid_cache': 'failed_obsids.yaml',
                'write_relpath': True
            },
        },
        'g3tsmurf': g3tsmurf_hwp_config.yaml',
        'imprinter': imprinter.yaml,
    }
"""

import traceback
import os
import argparse
import yaml
from dataclasses import dataclass, astuple, fields
import numpy as np
from tqdm.auto import tqdm
import logging

from typing import Optional, Union

from sotodlib import core
from sotodlib.io.metadata import write_dataset
from sotodlib.io.load_smurf import G3tSmurf, TuneSets
from sotodlib.io.load_book import load_smurf_npy_data, get_cal_obsids
from sotodlib.io.imprinter import Imprinter
import sotodlib.site_pipeline.util as sp_util

# stolen  from pysmurf, max bias volt / num_bits
DEFAULT_RTM_BIT_TO_VOLT = 10 / 2**19
DEFAULT_pA_per_phi0 = 9e6
TES_BIAS_COUNT = 12  # per detset / primary file group

logger = logging.getLogger('smurf_caldbs')
if not logger.hasHandlers():
    sp_util.init_logger('smurf_caldbs')



def main(config: Union[str, dict], 
        overwrite:Optional[bool]=False,
        skip_detset=False, skip_detcal=False):
    if not skip_detset:
        smurf_detset_info(config, overwrite)
    if not skip_detcal:
        run_update_det_caldb(config, overwrite=overwrite)

def smurf_detset_info(config: Union[str, dict], 
        overwrite:Optional[bool]=False):
    """Write out the updates for the manifest database with information about
    the readout ids present inside each detset.
    """
    if type(config) == str:
        config = yaml.safe_load(open(config, 'r'))

    h5_path = config['archive']['detset']['h5file']
    idx_path = config['archive']['detset']['index']
    root_dir = config['archive']['detset'].get('root_dir')
    if root_dir is not None:
        h5_path = os.path.join(root_dir, h5_path)
        idx_path = os.path.join(root_dir, idx_path)
    h5_relpath = os.path.relpath(h5_path, start=os.path.dirname(idx_path))

    config['g3tsmurf'] = os.path.expandvars(config['g3tsmurf'])
    config['imprinter'] = os.path.expandvars(config['imprinter'])

    SMURF = G3tSmurf.from_configs(config['g3tsmurf'])
    session = SMURF.Session()

    imprinter = Imprinter(config['imprinter'])
    ctx = core.Context(config['archive']['detset']['context'])

    if not os.path.exists(idx_path):
        scheme = core.metadata.ManifestScheme()
        scheme.add_exact_match('dets:detset')
        scheme.add_data_field('dataset')
        db = core.metadata.ManifestDb( 
            idx_path,
            scheme=scheme
        )
    else:
        db = core.metadata.ManifestDb(idx_path)

    keys = [
        "dets:readout_id",
        "dets:smurf.band",
        "dets:smurf.channel",
        "dets:smurf.frequency",
        "dets:stream_id",
        "dets:wafer_slot",
        "dets:tel_tube",
    ]

    stream_maps = {}
    for tube in imprinter.tubes:
        for s, slot in enumerate(imprinter.tubes[tube]['slots']):
            stream_maps[slot] = (f'ws{s}', tube)

    c = ctx.obsfiledb.conn.execute('select distinct name from detsets')
    ctx_detsets = [r[0] for r in c]
    added_detsets = db.get_entries(['dataset'])['dataset']
    detsets = session.query(TuneSets).all()

    for ts in detsets:
        if ts.name not in ctx_detsets:
            logger.debug(f"{ts.name} not in obsfiledb, ignoring")
            continue
        if ts.name in added_detsets and not overwrite:
            continue
        
        det_rs = core.metadata.ResultSet(keys=keys)
        for channel in ts.channels:
            det_rs.append({
                'dets:readout_id': channel.name,
                'dets:smurf.band': channel.band,
                'dets:smurf.channel': channel.channel,
                'dets:smurf.frequency': channel.frequency,
                'dets:stream_id': ts.stream_id,
                'dets:wafer_slot': stream_maps[ts.stream_id][0],
                'dets:tel_tube':stream_maps[ts.stream_id][1],
            })
        write_dataset(
            det_rs, h5_path,
            ts.name, 
            overwrite,
        )
        # add new entries to database
        write_relpath = config['archive']['detset'].get('write_relpath', True)
        if ts.name not in added_detsets:
            db_data = {'dets:detset': ts.name,
                    'dataset': ts.name}
            path = h5_relpath if write_relpath else h5_path
            db.add_entry(db_data, filename=path)


@dataclass
class CalInfo:
    """
    Class that contains detector calibration information that will go into the
    caldb.

    Attributes
    ------------
    readout_id: str
        Readout id of detector
    r_tes: float
        Detector resistance [ohms], determined through bias steps while the
        detector is biased
    r_frac: float
        Fractional resistance of TES, given by r_tes / r_n
    p_bias: float
        Bias power on the TES [J] computed using bias steps at the bias point
    s_i: float
        Current responsivity of the TES [1/V] computed using bias steps at the
        bias point
    phase_to_pW: float
        Phase to power conversion factor [pW/rad] computed using s_i,
        pA_per_phi0, and detector polarity
    v_bias: float
        Commanded bias voltage [V] on the bias line of the detector for the observation
    tau_eff: float
        Effective thermal time constant [sec] of the detector, measured from bias steps
    bg: int
        Bias group of the detector. Taken from IV curve data, which contains
        bgmap data taken immediately prior to IV. This will be -1 if the
        detector is unassigned
    polarity: int
        Polarity of the detector response for a positive change in bias current
        while the detector is superconducting. This is needed to correct for
        detectors that have reversed response. 
    r_n: float
        Normal resistance of the TES [Ohms] calculated from IV curve data
    p_sat: float
        "saturation power" of the TES [J] calculated from IV curve data.
        This is defined  as the electrical bias power at which the TES
        resistance is 90% of the normal resistance.
    """
    readout_id: str = ''

    # From bias steps
    r_tes: float = np.nan   # Ohm
    r_frac: float = np.nan
    p_bias: float = np.nan  # J
    s_i: float = np.nan     # 1/V
    phase_to_pW: float = np.nan  # pW/rad
    v_bias: float = np.nan  # V
    tau_eff: float = np.nan  # sec

    # From IV
    bg: int = -1
    polarity: int = 1
    r_n: float = np.nan  
    p_sat: float = np.nan   # J

    @classmethod
    def dtype(cls):
        """Returns ResultSet dtype for an item based on this class"""
        dtype = []
        for field in fields(cls):
            if field.name == 'readout_id':
                dt = ('dets:readout_id', '<U40')
            else:
                dt = (field.name, field.type)
            dtype.append(dt)
        return dtype


@dataclass
class CalResult:
    success: bool = False
    fail_msg: str = ''
    resset: Optional[core.metadata.ResultSet] = None


def get_cal_resset(ctx: core.Context, obs_id) -> CalResult:
    """
    Returns calibration ResultSet for a given ObsId. This pulls IV and bias step
    data for each detset in the observation, and uses that to compute CalInfo
    for each detector in the observation
    """
    am = ctx.get_obs(
            obs_id, samples=(0, 1), ignore_missing=True,
            no_signal=True, no_headers=False,
            on_missing={'det_cal': 'skip'}
    )
    cals = [CalInfo(rid) for rid in am.det_info.readout_id]

    if len(cals) == 0:
        return CalResult(success=False, fail_msg="No detectors present in observation.")

    if 'smurf' not in am.det_info:
        return CalResult(success=False, fail_msg="Missing smurf info for observation")

    iv_obsids = get_cal_obsids(ctx, obs_id, 'iv')

    rtm_bit_to_volt = None
    pA_per_phi0 = None
    ivas = {dset: None for dset in iv_obsids}
    for dset, oid in iv_obsids.items():
        if oid is not None:
            ivas[dset] = load_smurf_npy_data(ctx, oid, 'iv')
            if rtm_bit_to_volt is None:
                rtm_bit_to_volt = ivas[dset]['meta']['rtm_bit_to_volt']
                pA_per_phi0 = ivas[dset]['meta']['pA_per_phi0']
        else:
            logger.debug("missing IV data for %s", dset)

    bias_step_obsids = get_cal_obsids(ctx, obs_id, 'bias_steps')
    bsas = {dset: None for dset in bias_step_obsids}
    for dset, oid in bias_step_obsids.items():
        if oid is not None:
            bsas[dset] = load_smurf_npy_data(ctx, oid, 'bias_step_analysis')
            if rtm_bit_to_volt is None:
                rtm_bit_to_volt = bsas[dset]['meta']['rtm_bit_to_volt']
                pA_per_phi0 = ivas[dset]['meta']['pA_per_phi0']
        else:
            logger.debug("missing bias step data for %s", dset)

    if rtm_bit_to_volt is None:
        rtm_bit_to_volt = DEFAULT_RTM_BIT_TO_VOLT
    if pA_per_phi0 is None:
        pA_per_phi0 = DEFAULT_pA_per_phi0

    # Add IV info
    for i, cal in enumerate(cals):
        band = am.det_info.smurf.band[i]
        chan = am.det_info.smurf.channel[i]
        detset = am.det_info.detset[i]
        iva = ivas[detset]

        if iva is None: # No IV analysis for this detset
            continue

        ridx = np.where(
            (iva['bands'] == band) & (iva['channels'] == chan)
        )[0]
        if not ridx: # Channel doesn't exist in IV analysis
            continue

        ridx = ridx[0]
        cal.bg = iva['bgmap'][ridx]
        cal.polarity = iva['polarity'][ridx]
        cal.r_n = iva['R_n'][ridx]
        cal.p_sat = iva['p_sat'][ridx]

    obs_biases = dict(zip(am.bias_lines.vals, am.biases[:, 0] * 2*rtm_bit_to_volt))
    bias_line_is_valid = {k: True for k in obs_biases.keys()}

    # check to see if biases have changed between bias steps and obs
    for bsa in bsas.values():
        if bsa is None:
            continue

        for bg, vb_bsa in enumerate(bsa['Vbias']):
            bl_label = f"{bsa['meta']['stream_id']}_b{bg:0>2}"
            # Usually we can count on bias voltages of bias lines >= 12 to be
            # Nan, however we have seen cases where they're not, so we also
            # restrict by count.
            if np.isnan(vb_bsa) or bg >= TES_BIAS_COUNT:
                bias_line_is_valid[bl_label] = False
                continue

            if np.abs(vb_bsa - obs_biases[bl_label]) > 0.1:
                logger.debug("bias step and obs biases don't match for %s", bl_label)
                bias_line_is_valid[bl_label] = False

    for i, cal in enumerate(cals):
        band = am.det_info.smurf.band[i]
        chan = am.det_info.smurf.channel[i]
        detset = am.det_info.detset[i]
        stream_id = am.det_info.stream_id[i]
        bg = cal.bg
        bsa = bsas[detset]

        if bsa is None or bg == -1:
            continue

        bl_label = f'{stream_id}_b{bg:0>2}'
        if not bias_line_is_valid[bl_label]:
            continue

        ridx = np.where(
            (bsa['bands'] == band) & (bsa['channels'] == chan)
        )[0]
        if not ridx: # Channel doesn't exist in bias step analysis
            continue

        ridx = ridx[0]
        cal.r_tes = bsa['R0'][ridx]
        cal.r_frac = bsa['Rfrac'][ridx]
        cal.p_bias = bsa['Pj'][ridx]
        cal.s_i = bsa['Si'][ridx]
        cal.tau_eff = bsa['tau_eff'][ridx]
        if bg != -1:
            cal.v_bias = bsa['Vbias'][bg]
        cal.phase_to_pW = pA_per_phi0 / (2*np.pi) / cal.s_i * cal.polarity

    rset = core.metadata.ResultSet.from_friend(np.array(
        [astuple(c) for c in cals], dtype=CalInfo.dtype()
    ))
    return CalResult(success=True, resset=rset)


def get_obs_with_detsets(ctx, detset_idx):
    """Gets all observations with type 'obs' that have detset data"""
    db = core.metadata.ManifestDb(detset_idx)
    detsets = db.get_entries(['dataset'])['dataset']
    obs_ids = set()
    for dset in detsets:
        obs_ids = obs_ids.union(ctx.obsfiledb.get_obs_with_detset(dset))
    return obs_ids


def add_to_failed_cache(obs_id, cache_path, msg):
    logger.info(f"Adding {obs_id} to failed obsid cache with msg: {msg}")
    if os.path.exists(cache_path):
        with open(cache_path, 'r') as f:
            cache = yaml.safe_load(f)
    else:
        cache = {}

    cache[str(obs_id)] = msg
    with open(cache_path, 'w') as f:
        yaml.safe_dump(cache, f)

def get_failed_obsids(cache_path):
    if cache_path is None:
        return set([])
    if not os.path.exists(cache_path):
        return set([])
    with open(cache_path, 'r') as f:
        cache = yaml.safe_load(f)
    return set(cache.keys())

def update_det_caldb(ctx, idx_path, detset_idx, h5_path,
                     show_pb=False, format_exc=False, overwrite=False,
                     failed_obsid_cache=None, root_dir=None, write_relpath=True,
                     run_single=False):
    """
    Updates the detector caldb with new observations. This will find calibration
    data for all observations that have detset data and are not in the
    failed_obsid_cache if specified.

    Args
    -----
    ctx: core.Context
        Context object, must contain detset metadata
    idx_path: str
        Path to det_cal sqlite manifest db
    detset_idx: str
        Path to detset manifestdb
    h5_path: str
        Path to h5 file to write results set.
    show_pb: bool
        If True, will show progress bar and time remaining
    format_exc: bool
        If true, will log the full traceback whenever an exception is thrown.
    overwrite: bool
        If True, will overwrite existing entries in the h5 file.
    failed_obsid_cache: str
        Path to store failed obs-ids to avoid re-running them. If None, will not
        use a failed obsid cache and will attempt to add calibration info for all
        observations with detset data that are not in the det_cal manifest.
    root_dir: str
        Root directory of det_cal dbs. If true, will interpret ``idx_path``,
        ``h5_path``, and ``failed_obsid_cache`` relative to the root_dir.
    write_relpath: bool
        If true, when adding entries to the manifestdb, will use the h5 path relative
        to the idx_path.
    run_single: bool
        If true, will stop after writing a single entry to the det-cal db

    Returns
    -------
    report: dict
        Counts for how many items were 'success', 'failure', or 'crash'.

    """
    if root_dir is not None:
        h5_path = os.path.join(root_dir, h5_path)
        idx_path = os.path.join(root_dir, idx_path)
        if failed_obsid_cache is not None:
            failed_obsid_cache = os.path.join(root_dir, failed_obsid_cache)
    h5_relpath = os.path.relpath(h5_path, start=os.path.dirname(idx_path))

    if not os.path.exists(idx_path):
        scheme = core.metadata.ManifestScheme()
        scheme.add_exact_match('obs:obs_id')
        scheme.add_data_field('dataset')
        db = core.metadata.ManifestDb(scheme=scheme)
        db.to_file(idx_path)
    db = core.metadata.ManifestDb(idx_path)
    
    # detset_db = metadata.Manifest(detset_idx)
    obsids_with_detsets = get_obs_with_detsets(ctx, detset_idx)
    failed_obsids = get_failed_obsids(failed_obsid_cache)
    obsids_with_obs = set(ctx.obsdb.query("type=='obs'")['obs_id']) - failed_obsids

    remaining_obsids = obsids_with_detsets.intersection(obsids_with_obs)
    if not overwrite: 
        existing_obsids = set(db.get_entries(['dataset'])['dataset'])
        remaining_obsids = remaining_obsids - existing_obsids

    # Sort remaining obs_ids by timestamp
    remaining_obsids = sorted(remaining_obsids,
                              key=(lambda s: s.split('_')[1]))

    logger.info("%d datasets to add", len(remaining_obsids))
    report = {
        'success': 0,
        'failure': 0,
        'crash': 0,
    }
    for obs_id in tqdm(remaining_obsids, disable=(not show_pb)):
        try:
            result = get_cal_resset(ctx, obs_id)
            if result.success:
                logger.info("Writing metadata for %s", obs_id)
                write_dataset(result.resset, h5_path, obs_id, overwrite=overwrite)
                path = h5_relpath if write_relpath else h5_path
                db.add_entry({
                    'obs:obs_id': obs_id,
                    'dataset': obs_id,
                }, filename=path, replace=overwrite)
                report['success'] += 1
            else:
                logger.error("Failed on %s: %s", obs_id, result.fail_msg)
                if failed_obsid_cache is not None:
                    add_to_failed_cache(obs_id, failed_obsid_cache, result.fail_msg)
                report['failure'] += 1

        except Exception as e:
            logger.error("Crashed on %s: %s", obs_id, e)
            if format_exc:
                logger.error(traceback.format_exc())
            report['crash'] += 1

        if run_single:
            break

    return report

def run_update_det_caldb(config_path, overwrite=False):
    with open(config_path, 'r') as f:
        config = yaml.safe_load(f)

    format_exc = config['archive']['det_cal'].get('format_exc', False)
    detset_idx = config['archive']['detset']['index']
    detset_root_path = config['archive']['detset'].get('root_dir')
    if detset_root_path is not None:
        detset_idx = os.path.join(detset_root_path, detset_idx)

    ctx = core.Context(config['archive']['det_cal']['context'])
    report = update_det_caldb(
        ctx, 
        config['archive']['det_cal']['index'],
        detset_idx,
        config['archive']['det_cal']['h5file'],
        show_pb=config['archive']['det_cal'].get('show_pb', False),
        format_exc=format_exc,
        overwrite=overwrite,
        failed_obsid_cache=config['archive']['det_cal'].get('failed_obsid_cache'),
        root_dir=config['archive']['det_cal'].get('root_dir'),
        write_relpath=config['archive']['det_cal'].get('write_relpath', True),
        run_single=config['archive']['det_cal'].get('run_single', False)
    )

    logger.info(f'update_det_caldb summary: {report}')
    assert (report['crash'] == 0), "Unexpected data processing issues."


def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument(
        '--config', type=str, help="configuration file"
    )
    parser.add_argument(
        '--skip-detset', action='store_true', help="Skip detset update"
    )
    parser.add_argument(
        '--skip-detcal', action='store_true', help="Skip detcal update"
    )
    parser.add_argument(
        '--overwrite', action='store_true',
        help='Overwrite existing entries'
    )
    return parser


if __name__ == "__main__":
    parser = get_parser(parser=None)
    args = parser.parse_args()
    main(**vars(args))




"""In order to plug in nicely to the command line wrapper, element
submodules should expose functions called ``main()`` and
``get_parser()``.

The ``get_parser()`` function should look like this::

  def get_parser(parser=None):
    if parser is None:
      parser = argparse.ArgumentParser()
    # element-specific args:
    parser.add_argument('obs_id', help="The obs_id to analyze.")
    parser.add_arugment('--config-file', help="Config file.")
    return parser

When called by the ``so-site-pipeline`` wrapper, a parser will be
passed in.

The ``main()`` function is the entry point to be called from the CLI
or from Prefect.  The arguments should include the arguments defined
through the ArgumentParser, as well as any additional support for
Prefect (such as a logger argument).  For example::

  def main(obs_id=None, config_file=None, logger=None):
    ...


(Alternately a special entry point, cli_main, can be defined and that
will be used instead.)

If you want the submodule to be executable directly as a script (or
through ``python -m``), add a ``__main__`` handling block like this
one::

  if __name__ == '__main__':
    util.main_launcher(main, get_parser)


To register a properly organized submodule in the ``so-site-pipeline``
command line wrapper, edit ``cli.py`` and see comments inline.

"""

import argparse

from . import (
    analyze_bright_ptsrc,
    check_book,
    make_atomic_filterbin_map,
    make_det_info_wafer,
    make_ml_map,
    make_source_flags,
    make_uncal_beam_map,
    preprocess_tod,
    update_g3tsmurf_db,
    update_hkdb,
    update_obsdb,
    make_cosamp_hk
)

# Dictionary matching element name to a submodule (which must have
# been imported above).  Note for these to work, the submodule must
# have defined a get_parser() and a main() or cli_main() function as
# described in the module comments above.

ELEMENTS = {
    'analyze-bright-ptsrc': analyze_bright_ptsrc,
    'check-book': check_book,
    'make-atomic-filterbin-map': make_atomic_filterbin_map,
    'make-det-info-wafer': make_det_info_wafer,
    'make-ml-map': make_ml_map,
    'make-source-flags': make_source_flags,
    'make-uncal-beam-map': make_uncal_beam_map,
    'preprocess-tod': preprocess_tod,
    'update-g3tsmurf-db': update_g3tsmurf_db,
    'update-hkdb': update_hkdb,
    'update-obsdb': update_obsdb,
    'make-cosamp-hk': make_cosamp_hk,
}

CLI_NAME = 'so-site-pipeline'

def get_parser():
    parser = argparse.ArgumentParser()
    # Make sure all args here are redirected to vars starting with
    # '_'.  We are going to clean those off before passing to the
    # subcommand.
    parser.add_argument('--bash-completion', action='store_true',
                        dest='_bash_completion')
    sps = parser.add_subparsers(dest='_pipemod')

    for name, module in ELEMENTS.items():
        sp = sps.add_parser(name)
        module.get_parser(sp)

    return parser

def main():
    parser = get_parser()
    args = parser.parse_args()

    # Extract top-level args ...
    top_args = {k: v for k, v in vars(args).items()
                if k[0] == '_'}
    [delattr(args, k) for k in top_args]
    top_args = argparse.Namespace(**top_args)

    if top_args._bash_completion:
        names = list(ELEMENTS.keys())
        print('complete -W "%s" %s' % (' '.join(names), CLI_NAME))
        parser.exit()
    
    if top_args._pipemod is None:
        parser.error('First argument must be a sub-command.')

    module = ELEMENTS[top_args._pipemod]
    for epoint in ['cli_main', 'main']:
        _main = getattr(module, epoint, None)
        if _main is not None:
            _main(**vars(args))
            break
    else:
        print('No entry-point function found!')



from sotodlib import core, hwp, coords
import so3g
import numpy as np
from pixell import enmap
from ._helpers import quick_tod

import unittest

class DemodMapmakingTest(unittest.TestCase):
    def test_00_direct(self):
        """Test the coords.demod.make_map function on directly generated
        demodQ and demodU timestreams.

        """
        # Constant sky components.
        Q_stream = 1.0
        U_stream = 0.25
        TOL = 0.0001

        tod = quick_tod(10, 10000)
        fp  = coords.helpers.get_fplane(tod)
        csl = so3g.proj.CelestialSightLine.az_el(
            tod.timestamps, tod.boresight.az, tod.boresight.el, roll=tod.boresight.roll,
            site='so_sat1', weather='toco')

        for k in ['dsT', 'demodQ', 'demodU']:
            tod.wrap_new(k, shape=('dets', 'samps'), dtype='float32')

        for i, qdet in enumerate(tod.dets.vals):
            q_total = csl.Q * fp.quats[i]
            ra, dec, alpha = so3g.proj.quat.decompose_lonlat(q_total)
            GAMMA = alpha - 2 * tod.focal_plane.gamma[i]
            tod.demodQ[i] = Q_stream * np.cos(2 * GAMMA) + U_stream * np.sin(2 * GAMMA)
            tod.demodU[i] = U_stream * np.cos(2 * GAMMA) - Q_stream * np.sin(2 * GAMMA)

        results = coords.demod.make_map(tod)
        m0 = results['map']
        s = m0[1] != 0
        means = [m[s].mean() for m in m0]
        print(means)
        assert(abs(means[1] - Q_stream) < TOL)
        assert(abs(means[2] - U_stream) < TOL)

    def test_10_mod_demod(self):
        """Test the coords.demod.make_map function on timestreams starting
        from the HWP-modulated signal.

        """
        # Constant sky components.
        Q_stream = 1.0
        U_stream = 0.25
        TOL = .01

        tod = quick_tod(10, 10000)
        fp  = coords.helpers.get_fplane(tod)
        csl = so3g.proj.CelestialSightLine.az_el(
            tod.timestamps, tod.boresight.az, tod.boresight.el, roll=tod.boresight.roll,
            site='so_sat1', weather='toco')

        c_4chi, s_4chi = np.cos(tod.hwp_angle * 4), np.sin(tod.hwp_angle * 4)
        for i, det in enumerate(tod.dets.vals):
            q_total = csl.Q * fp.quats[i]
            ra, dec, alpha = so3g.proj.quat.decompose_lonlat(q_total)
            GAMMA = alpha - 2 * tod.focal_plane.gamma[i]
            c, s = np.cos(2*GAMMA), np.sin(2*GAMMA)
            tod.signal[i] = (Q_stream * (c*c_4chi - s*s_4chi) +
                             U_stream * (s*c_4chi + c*s_4chi))

        hwp.demod_tod(tod)
        results = coords.demod.make_map(tod)
        m0 = results['map']
        s = m0[1] != 0
        means = [m[s].mean() for m in m0]
        print(means)
        assert(abs(means[1] - Q_stream) < TOL)
        assert(abs(means[2] - U_stream) < TOL)
        
    def test_from_map_demodulated(self):
        """Test the coords.demod.from_map function of demodulated signal.

        """
        tod = quick_tod(10, 10000)
        TOL = 0.0001
        
        shape, wcs = enmap.fullsky_geometry(res=0.5*coords.DEG)
        signal_map = enmap.zeros((3, *shape), wcs)
        T_stream, Q_stream, U_stream = 1., 0.25, 0.01
        signal_map[0] += T_stream
        signal_map[1] += Q_stream
        signal_map[2] += U_stream
        _ = coords.demod.from_map(tod, signal_map, modulated=False, wrap=True)
        
        results = coords.demod.make_map(tod)
        m0 = results['map']
        s = m0[1] != 0
        means = [m[s].mean() for m in m0]
        assert(abs(means[1] - Q_stream) < TOL)
        assert(abs(means[2] - U_stream) < TOL)
        
    def test_from_map_modulated(self):
        """Test the coords.demod.from_map function of modulated signal.

        """
        tod = quick_tod(10, 10000)
        tod.move('signal', None)
        TOL = .01

        shape, wcs = enmap.fullsky_geometry(res=0.5*coords.DEG)
        signal_map = enmap.zeros((3, *shape), wcs)

        T_stream, Q_stream, U_stream = 1., 0.25, 0.01
        signal_map[0] += T_stream
        signal_map[1] += Q_stream
        signal_map[2] += U_stream

        # from_map expects and outputs float64
        _ = coords.demod.from_map(tod, signal_map, modulated=True, wrap=True)
        hwp.demod_tod(tod)
        # make_map expects float32
        for signal in ['signal', 'dsT', 'demodQ', 'demodU']:
           tod[signal] = tod[signal].astype('float32')
        results = coords.demod.make_map(tod)

        m0 = results['map']
        s = m0[1] != 0.
        means = [m[s].mean() for m in m0]
        assert(abs(means[1] - Q_stream) < TOL)
        assert(abs(means[2] - U_stream) < TOL)



from __future__ import annotations
from argparse import ArgumentParser
from typing import Optional, Union, Callable
from dataclasses import dataclass
import time
import datetime as dt
import warnings
import os
import logging
import yaml
import traceback
import ephem

from sqlalchemy import create_engine, select
from sqlalchemy.orm import sessionmaker

import numpy as np
import sotodlib.site_pipeline.util as util
from sotodlib import coords, mapmaking
from sotodlib.core import Context
from sotodlib.io import hk_utils
from sotodlib.preprocess import preprocess_util
from sotodlib.utils.procs_pool import get_exec_env
from so3g.proj import coords as so3g_coords
from pixell import enmap, wcsutils, colors, memory
from pixell import utils as putils
from pixell.mpiutils import FakeCommunicator


@dataclass
class Cfg:
    """
    Class to configure make-atomic-filterbin-map
    
    Args
    --------
    context: str
        Path to context file
    preprocess_config: str
        Path to config file(s) to run the preprocessing pipeline.
        If 2 files, representing 2 layers of preprocessing, they
        should be separated by a comma.
    area: str
        WCS kernel for rectangular pixels. Filename to map/geometry
        or valid string for coords.get_wcs_kernel.
    nside: int
        Nside for HEALPIX pixels
    query: str
        Query, can be a file (list of obs_id) or selection string (will select only CMB scans by default)
    odir: str
        Output directory
    update_delay: float
        Number of days in the past to start obs list
    site: str
    hk_data_path: str
        Path to housekeeping data
    nproc: int
        Number of procs for the multiprocessing pool
    atomic_db: str
        Path to the atomic map database
    comps: str
        Components to map, only TQU implemented
    singlestream: bool
        Map without demodulation (e.g. with a static HWP)
    only_hits: bool
        Only create a hits map
    all_splits: bool
        If True, map all implemented splits
    det_in_out: bool
        Make focal plane split: inner vs outer detector
    det_left_right: bool
        Make focal plane split: left vs right detector
    det_upper_lower: bool
        Make focal plane split: upper vs lower detector
    scan_left_right: bool
        Make samples split: left-going vs right going scans
    ntod: int
        Run the first ntod observations in your query
    tods: str
        Run a specific obs
    nset: int
        Run the first nset wafers
    wafer: str
        Run a specific wafer
    freq: str
        Run a specific frequency band
    unit: str
        Unit of data. Default is K
    use_psd: bool
        True by default. Use white noise measured by PSD
        as the weights for mapmaking. Must be provided by
        the preprocessing
    wn_label: str
        Path where to find the white noise per det by the
        preprocessing
    apply_wobble: bool
        Correct wobble deflection. This requires
        aman.wobble_params metadata in the context
    center_at: str
    max_dets: int
    fixed_time: int
    min_dur: int
    verbose: int
    quiet: int
    window: float
    dtype_tod: str
        Data type for timestreams
    dtype_map: str
        Data type for maps
    """
    def __init__(
        self,
        context: str,
        preprocess_config: str,
        area: Optional[str] = None,
        nside: Optional[int] = None,
        query: str = "type == 'obs' and subtype == 'cmb'",
        odir: str = "./output",
        update_delay: Optional[float] = None,
        site: str = 'so_sat3',
        hk_data_path: Optional[str] = None,
        nproc: int = 1,
        atomic_db: Optional[str] = None,
        comps: str = 'TQU',
        singlestream: bool = False,
        only_hits: bool = False,
        all_splits: bool = False,
        det_in_out: bool = False,
        det_left_right: bool = False,
        det_upper_lower: bool = False,
        scan_left_right: bool = False,
        ntod: Optional[int] = None,
        tods: Optional[str] = None,
        nset: Optional[int] = None,
        wafer: Optional[str] = None,
        freq: Optional[str] = None,
        center_at: Optional[str] = None,
        max_dets: Optional[int] = None,
        fixed_time: Optional[int] = None,
        min_dur: Optional[int] = None,
        verbose: int = 0,
        quiet: int = 0,
        window: Optional[float] = None,
        dtype_tod: str = 'float32',
        dtype_map: str = 'float64',
        unit: str = 'K',
        use_psd: bool = True,
        wn_label: str = 'preprocess.noiseQ_mapmaking.white_noise',
        apply_wobble: bool = True
    ) -> None:
        self.context = context
        self.preprocess_config = preprocess_config
        self.area = area
        self.nside = nside
        self.query = query
        self.odir = odir
        self.update_delay = update_delay
        self.site = site
        self.hk_data_path = hk_data_path
        self.nproc = nproc
        self.atomic_db = atomic_db
        self.comps = comps
        self.singlestream = singlestream
        self.only_hits = only_hits
        self.all_splits = all_splits
        self.det_in_out = det_in_out
        self.det_left_right = det_left_right
        self.det_upper_lower = det_upper_lower
        self.scan_left_right = scan_left_right
        self.ntod = ntod
        self.tods = tods
        self.nset = nset
        self.wafer = wafer
        self.freq = freq
        self.center_at = center_at
        self.max_dets = max_dets
        self.fixed_time = fixed_time
        self.min_dur = min_dur
        self.verbose = verbose
        self.quiet = quiet
        self.window = window
        self.dtype_tod = dtype_tod
        self.dtype_map = dtype_map
        self.unit = unit
        self.use_psd = use_psd
        self.wn_label = wn_label
        self.apply_wobble = apply_wobble
    @classmethod
    def from_yaml(cls, path) -> "Cfg":
        with open(path, "r") as f:
            d = yaml.safe_load(f)
            return cls(**d)

class DataMissing(Exception):
    pass

def get_pwv(start_time, stop_time, data_dir):
    try:
        pwv_info = hk_utils.get_hkaman(
            float(start_time), float(stop_time), alias=['pwv'],
            fields=['site.env-radiometer-class.feeds.pwvs.pwv'],
            data_dir=data_dir)
        pwv_all = pwv_info['env-radiometer-class']['env-radiometer-class'][0]
        pwv = np.nanmedian(pwv_all)
    except (KeyError, ValueError):
        pwv = 0.0
    return pwv

def get_sun_distance(site, ctime, az, el):
    site_ = so3g_coords.SITES[site].ephem_observer()
    dtime = dt.datetime.fromtimestamp(ctime, dt.timezone.utc)
    site_.date = ephem.Date(dtime)
    sun = ephem.Sun(site_)
    return np.degrees(ephem.separation((sun.az, sun.alt), (np.radians(az), np.radians(el))))

class ColoredFormatter(logging.Formatter):
    def __init__(self, msg, colors={'DEBUG':colors.reset,
                                    'INFO':colors.lgreen,
                                    'WARNING':colors.lbrown,
                                    'ERROR':colors.lred,
                                    'CRITICAL':colors.lpurple}):
        logging.Formatter.__init__(self, msg)
        self.colors = colors

    def format(self, record):
        try:
            col = self.colors[record.levelname]
        except KeyError:
            col = colors.reset
        return col + logging.Formatter.format(self, record) + colors.reset

class LogInfoFilter(logging.Filter):
    def __init__(self, rank=0):
        self.rank = rank
        try:
            # Try to get actual time since task start if possible
            import psutil
            p = psutil.Process(os.getpid())
            self.t0 = p.create_time()
        except ImportError:
            # Otherwise measure from creation of this filter
            self.t0 = time.time()

    def filter(self, record):
        record.rank = self.rank
        record.wtime = time.time() - self.t0
        record.wmins = record.wtime/60.
        record.whours = record.wmins/60.
        record.mem = memory.current()/1024.**3
        record.resmem = memory.resident()/1024.**3
        record.memmax = memory.max()/1024.**3
        return record


def future_write_to_log(e, errlog):
    errmsg = f'{type(e)}: {e}'
    tb = ''.join(traceback.format_tb(e.__traceback__))
    f = open(errlog, 'a')
    f.write(f'\n{time.time()}, future.result() error\n{errmsg}\n{tb}\n')
    f.close()

def main(
    config_file: str,
    executor: Union["MPICommExecutor", "ProcessPoolExecutor"],
    as_completed_callable: Callable) -> None:
    args = Cfg.from_yaml(config_file)

    # Set up logging.
    L = logging.getLogger(__name__)
    L.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    ch.setFormatter(ColoredFormatter(
        "%(wmins)7.2f %(mem)5.2f %(memmax)5.2f %(message)s"))
    ch.addFilter(LogInfoFilter())
    L.addHandler(ch)

    verbose = args.verbose - args.quiet

    recenter = None
    if args.center_at:
        recenter = mapmaking.parse_recentering(args.center_at)

    if args.area is not None: # Set shape, wcs for rectpix
        try:
            # Load shape, wcs from a map or geometry
            shape, wcs = enmap.read_map_geometry(args.area)
            wcs = wcsutils.WCS(wcs.to_header())
        except FileNotFoundError:
            # See if area is a wcs_kernel string
            try:
                shape = None
                wcs = coords.get_wcs_kernel(args.area)
            except ValueError:
                L.error("'area' not a valid filename or wcs_kernel string")
                exit(1)
        if recenter is not None:
            # wcs_kernel string not allowed for recenter=True
            if shape is None:
                L.error("'area' must be a map geometry file, not wcs_kernel string, to use recenter")
                exit(1)
        else:
            # If not recenter, we set shape=None so get_footprint will be called
            shape = None

    elif args.nside is not None:
        pass  # here I will map in healpix
    else:
        L.error('Neither rectangular area or nside specified, exiting.')
        exit(1)

    noise_model = mapmaking.NmatWhite()
    putils.mkdir(args.odir)

    preprocess_config_str = [s.strip() for s in args.preprocess_config.split(",")]
    preprocess_config = [] ; errlog = []
    for preproc_cf in preprocess_config_str:
        ff = open(preproc_cf, 'r')
        preproc_local = yaml.safe_load(ff)
        ff.close()
        preprocess_config.append( preproc_local )
        errlog.append( os.path.join(os.path.dirname(
            preproc_local['archive']['index']), 'errlog.txt') )


    if (args.update_delay is not None):
        min_ctime = int(time.time()) - args.update_delay*86400
        args.query += f" and timestamp>={min_ctime}"

    # Check for map data type
    if args.dtype_map == 'float32' or args.dtype_map == 'single':
        warnings.warn("You are using single precision for maps, we advice to use double precision")

    context_obj = Context(args.context)
    # obslists is a dict, obskeys is a list, periods is an array
    try:
        obslists, obskeys, periods, obs_infos = mapmaking.build_obslists(
            context_obj, args.query, nset=args.nset, wafer=args.wafer,
            freq=args.freq, ntod=args.ntod, tods=args.tods,
            fixed_time=args.fixed_time, mindur=args.min_dur)
    except mapmaking.NoTODFound as err:
        L.exception(err)
        exit(0)
    L.info(f'Running {len(obslists)} maps after build_obslists')

    split_labels = []
    if args.all_splits:
        raise ValueError('all_splits not implemented yet')
    if args.det_in_out:
        split_labels.append('det_in')
        split_labels.append('det_out')
    if args.det_left_right:
        split_labels.append('det_left')
        split_labels.append('det_right')
    if args.det_upper_lower:
        split_labels.append('det_upper')
        split_labels.append('det_lower')
    if args.scan_left_right:
        split_labels.append('scan_left')
        split_labels.append('scan_right')
    if not split_labels:
        split_labels.append('full')

    # We open the data base for checking if we have maps already,
    # if we do we will not run them again.
    if isinstance(args.atomic_db, str):
        if os.path.isfile(args.atomic_db) and not args.only_hits:
            engine = create_engine("sqlite:///%s" % args.atomic_db, echo=False)
            Session = sessionmaker(bind=engine)
            session = Session()

            keys_to_remove = []
            # Now we have obslists and splits ready, we look through the database
            # to remove the maps we already have from it
            for key, value in obslists.items():
                missing_split = False
                for split_label in split_labels:
                    query_ = select(mapmaking.AtomicInfo).filter_by(
                        obs_id=value[0][0],
                        telescope=obs_infos[value[0][3]].telescope,
                        freq_channel=key[2],
                        wafer=key[1],
                        split_label=split_label)
                    matches = session.execute(query_).scalars().all()
                    if len(matches) == 0:
                        # this means one of the requested splits is missing
                        # in the data base
                        missing_split = True
                        break
                if missing_split is False:
                    # this means we have all the splits we requested for the
                    # particular obs_id/telescope/freq/wafer
                    keys_to_remove.append(key)
            for key in keys_to_remove:
                obskeys.remove(key)
                del obslists[key]
            engine.dispose()

    obslists_arr = [item for key, item in obslists.items()]

    L.info(f'Running {len(obslists_arr)} maps after removing duplicate maps')

    # clean up lingering files from previous incomplete runs
    if len(preprocess_config)==1:
        policy_dir_init = os.path.join(os.path.dirname(preprocess_config[0]['archive']['policy']['filename']), 'temp')
    else:
        policy_dir_init = os.path.join(os.path.dirname(preprocess_config[0]['archive']['policy']['filename']), 'temp')
        policy_dir_proc = os.path.join(os.path.dirname(preprocess_config[1]['archive']['policy']['filename']), 'temp_proc')
    for obs in obslists_arr:
        obs_id = obs[0][0]
        if len(preprocess_config)==1:
            preprocess_util.cleanup_obs(obs_id, policy_dir_init, errlog[0], preprocess_config[0], subdir='temp', remove=False)
        else:
            preprocess_util.cleanup_obs(obs_id, policy_dir_init, errlog[0], preprocess_config[0], subdir='temp', remove=False)
            preprocess_util.cleanup_obs(obs_id, policy_dir_proc, errlog[1], preprocess_config[1], subdir='temp_proc', remove=False)

    # remove datasets from final archive file not found in db
    preprocess_util.cleanup_archive(preprocess_config[0], L)
    if len(preprocess_config) > 1:
        preprocess_util.cleanup_archive(preprocess_config[1], L)

    run_list = []
    for oi, ol in enumerate(obslists_arr):
        pid = ol[0][3]
        detset = ol[0][1]
        band = ol[0][2]
        obslist = ol
        t = putils.floor(periods[pid, 0])
        t5 = ("%05d" % t)[:5]
        prefix = "%s/%s/atomic_%010d_%s_%s" % (
            args.odir, t5, t, detset, band)

        tag = "%5d/%d" % (oi+1, len(obskeys))
        putils.mkdir(os.path.dirname(prefix))
        #pwv_atomic = get_pwv(periods[pid, 0], periods[pid, 1], args.hk_data_path)

        # Save file for data base of atomic maps.
        # We will write an individual file,
        # another script will loop over those files
        # and write into sqlite data base
        if not args.only_hits:
            info_list = []
            for split_label in split_labels:
                # info is a dictionary that is used to create an AtomicInfo object
                # to write into sqlite with sqlalchemy
                info = {}
                info['obs_id']=obslist[0][0]
                info['telescope']=obs_infos[obslist[0][3]].telescope
                info['freq_channel']=band
                info['wafer']=detset
                info['ctime']=int(t)
                info['split_label']=split_label
                info['split_detail'] = ''
                info['prefix_path'] = str(prefix + '_%s' % split_label)
                info['elevation'] = obs_infos[obslist[0][3]].el_center
                info['azimuth'] = obs_infos[obslist[0][3]].az_center
                #info['pwv'] = float(pwv_atomic)
                info['roll_angle'] = obs_infos[obslist[0][3]].roll_center
                info['sun_distance'] = get_sun_distance(args.site, int(t), obs_infos[obslist[0][3]].az_center, obs_infos[obslist[0][3]].el_center)
                info_list.append(info)
        # inputs that are unique per atomic map go into run_list
        if args.area is not None:
            run_list.append([obslist, shape, wcs, info_list, prefix, t, tag])
        elif args.nside is not None:
            run_list.append([obslist, None, None, info_list, prefix, t, tag])

    futures = [executor.submit(
            mapmaking.make_demod_map, args.context, r[0],
            noise_model, r[3], preprocess_config, r[4],
            shape=r[1], wcs=r[2], nside=args.nside,
            comm=FakeCommunicator(), t0=r[5], tag=r[6],
            recenter=recenter,
            dtype_map=args.dtype_map,
            dtype_tod=args.dtype_tod,
            comps=args.comps,
            verbose=verbose,
            split_labels=split_labels,
            singlestream=args.singlestream,
            site=args.site, unit=args.unit,
            use_psd=args.use_psd,
            wn_label=args.wn_label,apply_wobble=args.apply_wobble)
            for r in run_list]
    for future in as_completed_callable(futures):
        L.info('New future as_completed result')
        try:
            errors, outputs, d_ = future.result()
            if d_ is not None:
                list_infos = []
                for n_split in range(len(split_labels)):
                    list_infos.append(mapmaking.AtomicInfo.from_dict(d_[n_split]))
                mapmaking.atomic_db_aux(args.atomic_db, list_infos)
        except Exception as e:
            future_write_to_log(e, errlog[-1])
            continue
        futures.remove(future)
        for ii in range(len(errors)):
            for idx_prepoc in range(len(preprocess_config)):
                if isinstance(outputs[ii][idx_prepoc], dict):
                    preprocess_util.cleanup_mandb(errors[ii], outputs[ii][idx_prepoc], preprocess_config[idx_prepoc], L)
    L.info("Done")
    return True


def get_parser(parser: Optional[ArgumentParser] = None) -> ArgumentParser:
    if parser is None:
        p = ArgumentParser()
    else:
        p = parser
    p.add_argument(
        "--config_file", type=str, help="yaml file with configuration."
    )
    p.add_argument(
        "--nprocs", type=int, help="Number of processors to use."
        )
    return p

if __name__ == '__main__':
    args = get_parser().parse_args()
    rank, executor, as_completed_callable = get_exec_env(args.nprocs)
    if rank == 0:
        main(args.config_file, executor, as_completed_callable)



from __future__ import annotations
from typing import Tuple, Union, List, Callable, Optional


def _get_mpi_comm() -> (
    Tuple[bool, int, Optional["MPICommExecutor"], Optional[Callable]]
):
    """This private function tries to create an MPICommExecutor object and returns it if successful.

    Returns
    -------
    bool
        Whether an MPICommExecutor object was created successfully. This is important as only rank 0 creates the executor and all other ranks return None as an executor.
    int
        The global rank of the master process.
    Optional["MPICommExecutor"]
        If the executor was created successfully, this is the MPICommExecutor object. Otherwise, it is None.
    Optional[Callable]
        If the executor was created successfully, this is the as_completed function. Otherwise, it is None.
    """
    try:
        from mpi4py.futures import MPICommExecutor
        from mpi4py.futures import as_completed
        from mpi4py import MPI

        comm = MPI.COMM_WORLD

        rank = comm.Get_rank()
        max_workers = comm.Get_size() - 1

        return (
            True,
            rank,
            MPICommExecutor(comm, root=0, max_workers=max_workers).__enter__(),
            as_completed,
        )
    except (NameError, ImportError, ValueError):
        return False, 0, None, None


def _get_concurrent_comm(
    nprocs: Optional[int] = None,
) -> Tuple[int, "ProcessPoolExecutor", Callable]:
    """This private function tries to create an ProcessPoolExecutor object and returns it if successful.

    Returns
    -------
    bool
        Whether an ProcessPoolExecutor object was created successfully. This is important as only rank 0 creates the executor and all other ranks return None as an executor.
    int
        The global rank of the master process.
    Optional["ProcessPoolExecutor"]
        If the executor was created successfully, this is the ProcessPoolExecutor object. Otherwise, it is None.
    Optional[Callable]
        If the executor was created successfully, this is the as_completed function. Otherwise, it is None.
    """

    from concurrent.futures import ProcessPoolExecutor
    from concurrent.futures import as_completed
    from multiprocessing import set_start_method

    set_start_method("spawn")
    return True, 0, ProcessPoolExecutor(max_workers=nprocs), as_completed


def get_exec_env(
    nprocs: int = None, priority: List[str] = ["mpi", "process_pool"]
) -> Tuple[int, Union["MPICommExecutor", "ProcessPoolExecutor"], Callable]:
    """This function sets up the execution environment for parallel processing based on the specified priority list.

    Since this function returns a rank value the main function needs to go under an
    >>> if rank == 0

    block to ensure that the code is only run by the master process.

    In addition the main function needs to take the executor and as_completed_callable as arguments.

    Let's assume that the main function is called main_func and it current state is as follows:
    >>> def main_func():
    >>>     with ProcessPoolExecutor() as executor:
    >>>         futures = []
    >>>         for i in range(10):
    >>>             futures.append(executor.submit(some_func, i))
    >>>         for future in as_completed(futures):
    >>>             result = future.result()
    >>>             print(result)
    >>> if __name__ == "__main__":
    >>>     main_func()


    Utilizing different executors requires to change the main function as follows:
    >>> def main_func(executor, as_completed_callable):
    >>>     futures = []
    >>>     for i in range(10):
    >>>         futures.append(executor.submit(some_func, i))
    >>>     for future in as_completed_callable(futures):
    >>>         result = future.result()
    >>>         print(result)
    >>> if __name__ == "__main__":
    >>>     rank, executor, as_completed_callable = get_exec_env(nprocs=4)
    >>>     if rank == 0:
    >>>         main_func(executor, as_completed_callable)


    Parameters
    ----------
    nprocs : int, optional
        The number of processes to use for the process pool executor. If not specified, the default is None.
    priority : List[str], optional

    Returns
    -------
    int
        The rank of the process that is the master process for this type of execution

    Union["MPICommExecutor", "ProcessPoolExecutor"]
        The executor that is used to run the code in parallel. This can be either an MPICommExecutor or a ProcessPoolExecutor depending on the execution environment.

    Callable
        The as_completed function that is used to get the results of the parallel execution based on the type of the executor

    Raises
    ------
    ValueError
        When a valid executor is not available based on the execution priority list.
    """
    user_priority = list(priority)
    executor_created = False
    while not executor_created:
        executor_mode = user_priority.pop(0)
        if executor_mode == "mpi":
            executor_created, rank, executor, as_completed_callable = (
                _get_mpi_comm()
            )
        elif executor_mode == "process_pool":
            executor_created, rank, executor, as_completed_callable = (
                _get_concurrent_comm(nprocs)
            )
        else:
            raise ValueError("No executor available")

    return rank, executor, as_completed_callable



# Copyright (c) 2019-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import os
import sys
from datetime import datetime, timezone
import re

import numpy as np

import astropy.units as u
from astropy.table import QTable
from toast.instrument import Focalplane, GroundSite, Telescope
from toast.utils import Logger, name_UID

from ..core.hardware import Hardware, build_readout_id, parse_readout_id
from ..sim_hardware import sim_nominal
from .sim_focalplane import sim_telescope_detectors
from so3g.proj.coords import SITES


FOCALPLANE_RADII = {
    "LAT": u.Quantity(3.6, u.deg),
    "SAT1": u.Quantity(17.8, u.deg),
    "SAT2": u.Quantity(17.8, u.deg),
    "SAT3": u.Quantity(17.8, u.deg),
    "SAT4": u.Quantity(17.2, u.deg),
}


def get_telescope(telescope, wafer_slots, tube_slots):
    """Determine which telescope matches the detector selections"""
    if telescope is not None:
        return telescope
    # User did not set telescope so we infer it from the
    # tube and wafer slots
    hwexample = sim_nominal()
    if wafer_slots is not None:
        wafer_slots = wafer_slots.split(",")
        wafer_map = hwexample.wafer_map()
        tube_slots = [wafer_map["tube_slots"][ws] for ws in wafer_slots]
    elif tube_slots is not None:
        tube_slots = tube_slots.split(",")
    else:
        raise RuntimeError("Must set telescope, wafer_slots or tube_slots.")
    for tube_slot in tube_slots:
        for telescope_name, telescope_data in hwexample.data["telescopes"].items():
            if tube_slot in telescope_data["tube_slots"]:
                if telescope is None:
                    telescope = telescope_name
                elif telescope != telescope_name:
                    raise RuntimeError(
                        f"Tubes '{tube_slots}' span more than one telescope"
                    )
        if telescope is None:
            raise RuntimeError(
                f"Failed to match tube_slot = '{tube_slot}' with a telescope"
            )
    return telescope


def get_tele_wafer_band_name(tele, tube, wafer, band):
    """Return a simplified name for one band of a wafer.

    There are multiple places where we need to construct a unique string
    composed of the telescope name, optics tube, wafer and band within a
    wafer.  The formal band name includes the telescope, which results
    in annoying duplication.

    Args:
        tele (str):  Name of the telescope
        tube (str):  The optics tube slot
        wafer (str):  The wafer slot name
        band (str):  The name of the band in the hardware dictionary

    Returns:
        (str):  The name.

    """
    bf_pat = re.compile(r"(.*_)")
    band_freq = bf_pat.sub("", band)
    return f"{tele}_{tube}_{wafer}_{band_freq}"


class SOSite(GroundSite):
    def __init__(
        self,
        name="ATACAMA",
        lat=SITES["so"].lat * u.degree,
        lon=SITES["so"].lon * u.degree,
        alt=SITES["so"].elev * u.meter,
        weather="atacama",
        **kwargs,
    ):
        super().__init__(
            name,
            lat,
            lon,
            alt,
            weather=weather,
            **kwargs,
        )


class SOFocalplane(Focalplane):
    """SO Focalplane class.

    This can be constructed in several ways:
        - From a file while applying selection criteria.
        - From an existing (pre-selected) Hardware instance.
        - From a nominal sim on the fly with selections.

    Note that to support serialization, the simulated hardware dictionary elements
    have values in standard units.  When constructing our focalplane table we restore
    those units and build Quantities.

    Args:
        hw (Hardware):  If specified, construct from a Hardware object in memory.
        hwfile (str):  If specified, load this hardware model from disk.
        det_info_file (str):  If simulating a Hardware model, optionally specify
            a det_info format file to load
        det_info_version (str):  The version of the det_info file format.
        telescope (str):  If not None, select only detectors from this telescope.
        sample_rate (Quantity):  Use this sample rate for all detectors.
        bands (str):  Comma separated string of bands to use.
        wafer_slots (str):  Comma separated string of wafers to use.
        tube_slots (str):  Comma separated string of tubes to use.
        thinfp (int):  The factor by which to reduce the number of detectors.
        creation_time (float):  Optional timestamp to use when building readout_id.
        comm (MPI.Comm):  Optional MPI communicator.
        apply_net_corr (bool):  Degrade NETs according to the correlation factors.
            If False, the correlation factors are just recorded in the instrument
            model.

    """

    def __init__(
        self,
        hw=None,
        hwfile=None,
        det_info_file=None,
        det_info_version=None,
        telescope=None,
        sample_rate=u.Quantity(10.0, u.Hz),
        bands=None,
        wafer_slots=None,
        tube_slots=None,
        thinfp=None,
        creation_time=None,
        comm=None,
        apply_net_corr=True,
    ):
        log = Logger.get()
        meta = dict()
        meta["telescope"] = get_telescope(telescope, wafer_slots, tube_slots)
        field_of_view = 2 * FOCALPLANE_RADII[meta["telescope"]]

        if creation_time is None:
            # Use the current time
            creation_time = datetime.now(tz=timezone.utc).timestamp()

        if hw is None:
            if comm is None or comm.rank == 0:
                if hwfile is not None:
                    log.debug(f"Loading hardware configuration from {hwfile}...")
                    hw = Hardware(hwfile)
                elif meta["telescope"] in ["LAT", "SAT1", "SAT2", "SAT3", "SAT4"]:
                    log.debug("Simulating default hardware configuration")
                    hw = sim_nominal()
                    sim_telescope_detectors(
                        hw,
                        meta["telescope"],
                        det_info=(det_info_file, det_info_version),
                        no_darks=det_info_file is not None,
                    )
                else:
                    raise RuntimeError(
                        "Must provide a path to file or a valid telescope name"
                    )

                if (
                    bands is not None
                    or wafer_slots is not None
                    or tube_slots is not None
                ):
                    match = dict()
                    if bands is not None:
                        match["band"] = bands.replace(",", "|")
                    if wafer_slots is not None:
                        match["wafer_slot"] = wafer_slots.split(",")
                    if tube_slots is not None:
                        tube_slots = tube_slots.split(",")
                    hw = hw.select(tube_slots=tube_slots, match=match)

                if thinfp is not None:
                    dets = list(hw.data["detectors"].keys())
                    for det in dets:
                        pixel = hw.data["detectors"][det]["pixel"]
                        try:
                            pixel_id = int(pixel)
                        except ValueError:
                            pixel_id = name_UID(pixel)
                        if pixel_id % thinfp != 0:
                            del hw.data["detectors"][det]

                ndet = len(hw.data["detectors"])
                if ndet == 0:
                    raise RuntimeError(
                        f"No detectors match query: telescope={meta['telescope']}, "
                        f"tube_slots={tube_slots}, wafer_slots={wafer_slots}, "
                        f"bands={bands}, thinfp={thinfp}"
                    )
                else:
                    log.debug(
                        f"{ndet} detectors match query: telescope={meta['telescope']}, "
                        f"tube_slots={tube_slots}, wafer_slots={wafer_slots}, "
                        f"bands={bands}, thinfp={thinfp}"
                    )

            if comm is not None:
                hw = comm.bcast(hw)

        def get_par_float(ddata, key, default):
            if key in ddata:
                return float(ddata[key])
            else:
                return float(default)

        (
            readout_id,
            names,
            quats,
            bands,
            wafer_slots,
            tube_slots,
            nets,
            net_corrs,
            fknees,
            fmins,
            alphas,
            As,
            Cs,
            bandcenters,
            bandwidths,
            ids,
            pixels,
            fwhms,
            pols,
            pol_angs,
            pol_angs_wafer,
            pol_orientations_wafer,
            gamma,
            wafer_xs,
            wafer_ys,
            card_slots,
            channels,
            AMCs,
            biases,
            readout_freqs,
            bondpads,
            mux_positions,
            tele_wf_band,
        ) = (
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
            [],
        )

        for det_name, det_data in hw.data["detectors"].items():
            readout_id.append(
                build_readout_id(
                    creation_time, det_data["wafer_slot"], det_data["channel"]
                )
            )
            names.append(det_name)
            quats.append(np.array([float(x) for x in det_data["quat"]]))
            ids.append(int(det_data["ID"]))
            # Replace pixel index on the wafer with a unique identifier that
            # combines telescope, wafer and pixel index
            pixels.append(
                f"{meta['telescope']}_{det_data['wafer_slot']}_p{det_data['pixel']}"
            )
            fwhms.append(float(det_data["fwhm"]) * u.arcmin)
            pols.append(det_data["pol"])
            pol_angs.append(det_data["pol_ang"] * u.degree)
            pol_angs_wafer.append(det_data["pol_ang_wafer"] * u.degree)
            pol_orientations_wafer.append(det_data["pol_orientation_wafer"] * u.degree)
            gamma.append(det_data["pol_ang"] * u.degree)
            wafer_xs.append(det_data["wafer_x_mm"] * u.mm)
            wafer_ys.append(det_data["wafer_y_mm"] * u.mm)
            card_slots.append(det_data["card_slot"])
            channels.append(det_data["channel"])
            AMCs.append(det_data["AMC"])
            biases.append(det_data["bias"])
            readout_freqs.append(float(det_data["readout_freq_GHz"]) * u.GHz)
            bondpads.append(det_data["bondpad"])
            mux_positions.append(det_data["mux_position"])
            # Band is used to retrieve band-averaged values
            band = det_data["band"]
            band_data = hw.data["bands"][band]
            bands.append(band)
            # Get the wafer slot and translate into tube slot
            wafer_slot = det_data["wafer_slot"]
            wafer_slots.append(wafer_slot)
            # Determine which tube_slot has this wafer
            for tube_slot, tube_data in hw.data["tube_slots"].items():
                if wafer_slot in tube_data["wafer_slots"]:
                    break
            else:
                raise RuntimeError(f"{wafer_slot} is not in any tube_slot")
            tube_slots.append(tube_slot)
            tele_wf_band.append(
                get_tele_wafer_band_name(meta["telescope"], tube_slot, wafer_slot, band)
            )
            # Get noise parameters.  If detector-specific entries are
            # absent, use band averages
            net_corr = get_par_float(
                det_data, "NET_corr", band_data["NET_corr"]
            )
            net = get_par_float(det_data, "NET", band_data["NET"]) \
                * 1.0e-6 * u.K * u.s**0.5
            if apply_net_corr:
                net *= net_corr
                net_corr = 1.0
            nets.append(net)
            net_corrs.append(net_corr)
            fknees.append(get_par_float(det_data, "fknee", band_data["fknee"]) * u.mHz)
            fmins.append(get_par_float(det_data, "fmin", band_data["fmin"]) * u.mHz)
            alphas.append(get_par_float(det_data, "alpha", band_data["alpha"]))
            As.append(get_par_float(det_data, "A", band_data["A"]))
            Cs.append(get_par_float(det_data, "C", band_data["C"]))
            # bandpass
            lower = get_par_float(det_data, "low", band_data["low"]) * u.GHz
            # center = get_par_float(det_data, "center", band_data["center"]) * u.GHz
            upper = get_par_float(det_data, "high", band_data["high"]) * u.GHz
            bandcenters.append(0.5 * (lower + upper))
            bandwidths.append(upper - lower)

        meta["platescale"] = hw.data["telescopes"][meta["telescope"]]["platescale"] \
                             * u.deg / u.mm

        detdata = QTable(
            [
                readout_id,
                names,
                ids,
                quats,
                bands,
                card_slots,
                wafer_slots,
                tube_slots,
                fwhms,
                nets,
                net_corrs,
                fknees,
                fmins,
                alphas,
                As,
                Cs,
                bandcenters,
                bandwidths,
                pixels,
                pols,
                pol_angs,
                pol_angs_wafer,
                pol_orientations_wafer,
                gamma,
                wafer_xs,
                wafer_ys,
                channels,
                AMCs,
                biases,
                readout_freqs,
                bondpads,
                mux_positions,
                tele_wf_band,
            ],
            names=[
                "readout_id",
                "name",
                "uid",
                "quat",
                "band",
                "card_slot",
                "wafer_slot",
                "tube_slot",
                "FWHM",
                "psd_net",
                "NET_corr",
                "psd_fknee",
                "psd_fmin",
                "psd_alpha",
                "elevation_noise_a",
                "elevation_noise_c",
                "bandcenter",
                "bandwidth",
                "pixel",
                "pol",
                "pol_ang",
                "pol_ang_wafer",
                "pol_orientation_wafer",
                "gamma",
                "wafer_x",
                "wafer_y",
                "channel",
                "AMC",
                "bias",
                "readout_freq",
                "bondpad",
                "mux_position",
                "tele_wf_band",
            ],
            meta=meta,
        )

        super().__init__(
            detector_data=detdata,
            field_of_view=field_of_view,
            sample_rate=sample_rate,
        )


def update_creation_time(det_data, creation_time):
    """Update the readout_id column of a focalplane with a new creation time.

    Args:
        det_data (QTable):  The detector properties table.
        creation_time (float):  The updated time for use in the readout_id

    Returns:
        None

    """
    for row in range(len(det_data)):
        wf, ct, chan = parse_readout_id(det_data["readout_id"][row])
        det_data["readout_id"][row] = build_readout_id(creation_time, wf, chan)


def simulated_telescope(
    hw=None,
    hwfile=None,
    det_info_file=None,
    det_info_version=None,
    telescope_name=None,
    sample_rate=10 * u.Hz,
    bands=None,
    wafer_slots=None,
    tube_slots=None,
    thinfp=None,
    comm=None,
):
    if hw is not None and telescope_name is None:
        # get it from the hw
        if len(hw.data["telescopes"]) != 1:
            raise RuntimeError("Input Hardware has multiple telescopes")
        telescope_name = list(hw.data["telescopes"].keys())[0]
    focalplane = SOFocalplane(
        hw=hw,
        hwfile=hwfile,
        det_info_file=det_info_file,
        det_info_version=det_info_version,
        telescope=telescope_name,
        sample_rate=sample_rate,
        bands=bands,
        wafer_slots=wafer_slots,
        tube_slots=tube_slots,
        thinfp=thinfp,
        comm=comm,
    )
    site = SOSite()
    # The focalplane construction above will lookup the telescope name if needed
    # from the wafer / tube information
    telescope = Telescope(
        focalplane.detector_data.meta["telescope"],
        focalplane=focalplane,
        site=site,
    )
    return telescope



# Copyright (c) 2023-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Data I/O operations.
"""

import os
import numpy as np
from astropy import units as u
import toast
import toast.ops

from .. import ops as so_ops
from .job import workflow_timer


def setup_load_data_context(operators):
    """Add commandline args and operators for loading from a context.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(
        so_ops.LoadContext(
            name="load_context",
            enabled=False,
        )
    )
    return


@workflow_timer
def load_data_context(job, otherargs, runargs, data):
    """Load data from a Context.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    # Load it
    if job_ops.load_context.enabled:
        # Special handling of ACT data.  In this case, the user may have
        # set the MOBY2_TOD_STAGING_PATH environment variable to force the
        # data loader to make a copy into ramdisk for opening.  This will
        # break parallel access, since each process tries to create (and
        # then delete) the same file.  Here we append the process rank to
        # the staging directory
        if "MOBY2_TOD_STAGING_PATH" in os.environ:
            procdir = os.path.join(
                os.environ["MOBY2_TOD_STAGING_PATH"],
                f"{data.comm.world_rank}",
                f"{os.getpid()}",
            )
            os.makedirs(procdir, exist_ok=True)
            os.environ["MOBY2_TOD_STAGING_PATH"] = procdir
        job_ops.load_context.apply(data)


def setup_load_data_hdf5(operators):
    """Add commandline args and operators for loading from HDF5 files.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.LoadHDF5(name="load_hdf5", enabled=False))


@workflow_timer
def load_data_hdf5(job, otherargs, runargs, data):
    """Load data from one or more HDF5 files.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    # Load it
    if job_ops.load_hdf5.enabled:
        job_ops.load_hdf5.apply(data)


def setup_load_data_books(operators):
    """Add commandline args and operators for directly loading L3 books.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(so_ops.LoadBooks(name="load_books", enabled=False))


@workflow_timer
def load_data_books(job, otherargs, runargs, data):
    """Load data from one or more books.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    # Load it
    if job_ops.load_books.enabled:
        job_ops.load_books.apply(data)


def setup_save_data_hdf5(operators):
    """Add commandline args and operators for saving to HDF5 files.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.SaveHDF5(name="save_hdf5", enabled=False))


@workflow_timer
def save_data_hdf5(job, otherargs, runargs, data):
    """Save data to HDF5 files.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    # Dump it
    if job_ops.save_hdf5.enabled:
        if hasattr(otherargs, "out_dir"):
            hdf5_out = os.path.join(otherargs.out_dir, "data")
            job_ops.save_hdf5.volume = hdf5_out
        job_ops.save_hdf5.apply(data)


def setup_save_data_books(operators):
    """Add commandline args and operators for saving to L3 books.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(so_ops.SaveBooks(name="save_books", enabled=False))


@workflow_timer
def save_data_books(job, otherargs, runargs, data):
    """Save data to books.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    # Load it
    if job_ops.save_books.enabled:
        job_ops.save_books.apply(data)
